from __future__ import annotations

import datetime
import enum
import re
from typing import Any, Optional, Union, cast
from uuid import uuid4


class ValueObject:
    __slots__ = ("value",)

    def __init__(self, value: Any, /):
        self.value = value

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value == cast(ValueObject, other).value
        else:
            return NotImplemented

    def __ne__(self, other: object) -> bool:
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        else:
            return not result

    def __lt__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value < cast(ValueObject, other).value
        else:
            return NotImplemented

    def __le__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value <= cast(ValueObject, other).value
        else:
            return NotImplemented

    def __gt__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value > cast(ValueObject, other).value
        else:
            return NotImplemented

    def __ge__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value >= cast(ValueObject, other).value
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value!r})"

    def __str__(self) -> str:
        return str(self.value)


class Flag(ValueObject):
    def __init__(self, value: bool, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )

        if not isinstance(value, bool):
            raise TypeError(
                f"{self.__class__.__name__} expects an integer, but got {value!r}"
            )

        super().__init__(value)


class _PositiveInt(ValueObject):
    def __init__(self, value: int, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, int):
            if value < 0:
                raise ValueError(
                    f"{self.__class__.__name__} value must be positive, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects an integer, but got {value!r}"
            )

        super().__init__(value)


class _NonEmptyStr(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if not value:
                raise ValueError(
                    f"{self.__class__.__name__} value is required, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a string, but got {value!r}."
            )

        super().__init__(value)


class Success(ValueObject):
    def __init__(self) -> None:
        super().__init__(None)

    def __repr__(self) -> str:
        return "Success()"

    def __str__(self) -> str:
        return "Success"


class Failure(ValueObject):
    def __init__(self, message: str, /) -> None:
        value = _NonEmptyStr(message).value
        super().__init__(value)


class Result(ValueObject):
    def __init__(self, value: Union[Failure, Success], /) -> None:
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )

        if not isinstance(value, (Failure, Success)):
            raise TypeError(
                f"{self.__class__.__name__} expects either a Success or Failure instance, "
                f"but got {value!r}."
            )

        super().__init__(value)

    @classmethod
    def failure(cls, message: str, /) -> Result:
        return Result(Failure(message))

    @classmethod
    def success(cls) -> Result:
        return Result(Success())

    @property
    def failure_message(self) -> str:
        if self.is_failure:
            return self.value.value
        else:
            raise TypeError(
                f"{self.__class__.__name__} does not contain a failure value, so it does not have "
                f"a failure message.  The value of the Result is {self.value!r}."
            )

    @property
    def failure_message_or_none(self) -> Optional[str]:
        if self.is_failure:
            return self.value.value
        else:
            return None

    @property
    def is_failure(self) -> bool:
        return isinstance(self.value, Failure)

    @property
    def is_success(self) -> bool:
        return not self.is_failure


class UniqueId(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if len(value) != 32:
                raise ValueError(
                    f"{self.__class__.__name__} value must be 32 characters long, "
                    f"but got {value!r}."
                )
            if not value.isalnum():
                raise ValueError(
                    f"{self.__class__.__name__} value must be all alphanumeric characters, but "
                    f"got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)

    @classmethod
    def generate(cls) -> UniqueId:
        return UniqueId(uuid4().hex)


class SchemaName(ValueObject):
    def __init__(self, value: Optional[str], /):
        if value is None:
            ...
        elif isinstance(value, str):
            if not value:
                raise ValueError(
                    f"If a {self.__class__.__name__} value is provided, then it must be at least 1 "
                    f"character long, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class SingleChar(ValueObject):
    def __init__(self, value: str, /):
        if not value:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got {value!r}."
            )
        elif isinstance(value, str):
            if len(value) != 1:
                raise ValueError(
                    f"{self.__class__.__name__} must be 1 char long, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class JobName(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if len(value) < 3 or len(value) >= 100:
                raise ValueError(
                    f"{self.__class__.__name__} must be between 3 and 100 characters long, but got "
                    f"{value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class SecondsBetweenRefreshes(ValueObject):
    def __init__(self, value: int, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, int):
            if value < 0:
                raise ValueError(
                    f"If a {self.__class__.__name__} must be positive, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects an int, but got {value!r}"
            )

        super().__init__(value)


class TestName(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if len(value) < 3 or len(value) > 100:
                raise ValueError(
                    f"{self.__class__.__name__} must be between 3 and 100 characters long, "
                    f"but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class Days(_PositiveInt):
    ...


class DaysToKeep(_PositiveInt):
    ...


class ExecutionMillis(_PositiveInt):
    ...


class TimeoutSeconds(_PositiveInt):
    ...


class MaxRetries(_PositiveInt):
    ...


class FlexPercent(ValueObject):
    def __init__(self, value: float, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, float):
            if value < 0 or value > 1:
                raise ValueError(
                    f"{self.__class__.__name__} value must be between 0 and 1, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a float, but got {value!r}"
            )

        super().__init__(value)


class Timestamp(ValueObject):
    def __init__(self, value: datetime.datetime, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )

        if not isinstance(value, datetime.datetime):
            raise TypeError(
                f"{self.__class__.__name__} expects a datetime.datetime, but got {value!r}"
            )

        super().__init__(value)

    @classmethod
    def now(cls) -> Timestamp:
        return Timestamp(datetime.datetime.now())


class SMTPServer(_NonEmptyStr):
    ...


class SMTPPort(_PositiveInt):
    ...


class EmailAddress(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if (
                re.match(r"^[a-zA-Z0-9]+[._]?[a-zA-Z0-9]+[@]\w+[.]\w{2,3}$", value)
                is None
            ):
                raise ValueError(f"{value!r} is not a valid EmailAddress.")
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class EmailSubject(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if len(value) < 3 or len(value) > 200:
                raise ValueError(
                    f"{self.__class__.__name__} must be between 3 and 200 characters long, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class Password(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if not value.strip():
                raise ValueError(
                    f"{self.__class__.__name__} value is required, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)

    def __repr__(self) -> str:
        return f"Password({'*' * 10})"

    def __str__(self) -> str:
        return "*" * 10


class EmailMsg(_NonEmptyStr):
    ...


class LogLevelOption(enum.Enum):
    Debug = 1
    Info = 2
    Error = 3


class LogLevel(ValueObject):
    def __init__(self, value: LogLevelOption, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )

        if not isinstance(value, LogLevelOption):
            raise TypeError(
                f"{self.__class__.__name__} expects a LogLevelOption value, but got {value!r}."
            )

        super().__init__(value)

    @classmethod
    def debug(cls) -> LogLevel:
        return LogLevel(LogLevelOption.Debug)

    @classmethod
    def error(cls) -> LogLevel:
        return LogLevel(LogLevelOption.Error)

    @classmethod
    def info(cls) -> LogLevel:
        return LogLevel(LogLevelOption.Info)

    def __str__(self) -> str:
        if self.value == LogLevelOption.Debug:
            return "DEBUG"
        elif self.value == LogLevelOption.Info:
            return "INFO"
        elif self.value == LogLevelOption.Error:
            return "ERROR"
        else:
            raise ValueError(f"{self.value} is not a LogLevelOption value.")


class LogMessage(_NonEmptyStr):
    ...


class SecondsSinceLastRefresh(ValueObject):
    def __init__(self, value: Optional[int], /):
        if value is None:
            ...
        elif isinstance(value, int):
            if not value:
                raise ValueError(
                    f"If a {self.__class__.__name__} value is provided, then it must be positive, "
                    f"but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects an int, but got {value!r}"
            )

        super().__init__(value)



import abc


from lime_etl.domain import job_log_entry, value_objects
from lime_etl.services import unit_of_work


class JobLoggingService(abc.ABC):
    @abc.abstractmethod
    def log_error(
        self,
        message: value_objects.LogMessage,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def log_info(
        self,
        message: value_objects.LogMessage,
    ) -> None:
        raise NotImplementedError


class DefaultJobLoggingService(JobLoggingService):
    def __init__(
        self,
        uow: unit_of_work.UnitOfWork,
        batch_id: value_objects.UniqueId,
        job_id: value_objects.UniqueId,
    ):
        self._uow = uow
        self.batch_id = batch_id
        self.job_id = job_id

        super().__init__()

    def _log(
        self,
        level: value_objects.LogLevel,
        message: value_objects.LogMessage,
    ) -> None:
        with self._uow as uow:
            ts = uow.ts_adapter.now()
            log_entry = job_log_entry.JobLogEntry(
                id=value_objects.UniqueId.generate(),
                batch_id=self.batch_id,
                job_id=self.job_id,
                log_level=level,
                message=message,
                ts=ts,
            )
            print(log_entry)
            uow.job_log.add(log_entry=log_entry)
            uow.commit()
            return None

    def log_error(
        self,
        message: value_objects.LogMessage,
    ) -> None:
        return self._log(
            level=value_objects.LogLevel.error(),
            message=message,
        )

    def log_info(
        self,
        message: value_objects.LogMessage,
    ) -> None:
        return self._log(
            level=value_objects.LogLevel.info(),
            message=message,
        )


class ConsoleJobLoggingService(JobLoggingService):
    def __init__(self) -> None:
        super().__init__()

    def log_error(
        self,
        message: value_objects.LogMessage,
    ) -> None:
        print(f"ERROR: {message.value}")
        return None

    def log_info(
        self,
        message: value_objects.LogMessage,
    ) -> None:
        print(f"INFO: {message.value}")
        return None



from __future__ import annotations

import itertools
import typing

import sqlalchemy as sa
from sqlalchemy.orm import sessionmaker

from lime_etl.adapters import email_adapter, orm, timestamp_adapter
from lime_etl.domain import batch_delta, job_spec, value_objects
from lime_etl.services import batch_runner, unit_of_work
from lime_etl.services.admin import delete_old_logs

DEFAULT_ADMIN_JOBS = [
    delete_old_logs.DeleteOldLogs(days_to_keep=value_objects.DaysToKeep(3)),
]


def run(
    *,
    engine_or_uri: typing.Union[sa.engine.Engine, str],
    etl_jobs: typing.Iterable[job_spec.JobSpec],
    admin_jobs: typing.Iterable[job_spec.JobSpec] = frozenset(DEFAULT_ADMIN_JOBS),
    schema: typing.Optional[str] = None,
    ts_adapter: timestamp_adapter.TimestampAdapter = timestamp_adapter.LocalTimestampAdapter(),
    email_adapter: typing.Optional[email_adapter.EmailAdapter] = None
) -> batch_delta.BatchDelta:
    if schema:
        orm.set_schema(schema=value_objects.SchemaName(schema))

    orm.start_mappers()
    if type(engine_or_uri) is sa.engine.Engine:
        engine = engine_or_uri
    else:
        engine = sa.create_engine(engine_or_uri)
    orm.metadata.create_all(engine)
    session_factory = sessionmaker(bind=engine)
    uow = unit_of_work.DefaultUnitOfWork(session_factory=session_factory)
    result: batch_delta.BatchDelta = batch_runner.run(
        uow=uow, jobs=list(itertools.chain(admin_jobs, etl_jobs)), ts_adapter=ts_adapter
    )
    if email_adapter:
        email_adapter.send(result=result)
    return result



import abc
import datetime
from typing import List, Optional

from sqlalchemy import desc
from sqlalchemy.orm import Session

from lime_etl.adapters import timestamp_adapter
from lime_etl.domain import batch, job_result, value_objects


class BatchRepository(abc.ABC):
    @abc.abstractmethod
    def add(self, new_batch: batch.Batch) -> batch.Batch:
        raise NotImplementedError

    @abc.abstractmethod
    def add_job_result(self, result: job_result.JobResult) -> job_result.JobResult:
        raise NotImplementedError

    @abc.abstractmethod
    def delete_old_entries(self, days_to_keep: value_objects.DaysToKeep) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def get_batch_by_id(self, batch_id: value_objects.UniqueId) -> Optional[batch.Batch]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_latest(self) -> Optional[batch.Batch]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_latest_result_for_job(
        self, job_name: value_objects.JobName
    ) -> Optional[job_result.JobResult]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_last_successful_ts_for_job(
        self, job_name: value_objects.JobName
    ) -> Optional[value_objects.Timestamp]:
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, batch_to_update: batch.Batch) -> batch.Batch:
        raise NotImplementedError


class SqlAlchemyBatchRepository(BatchRepository):
    def __init__(
        self, session: Session, ts_adapter: timestamp_adapter.TimestampAdapter
    ):
        self._session = session
        self._ts_adapter = ts_adapter
        super().__init__()

    def add(self, new_batch: batch.Batch) -> batch.Batch:
        dto = new_batch.to_dto()
        self._session.add(dto)
        return new_batch

    def add_job_result(self, result: job_result.JobResult) -> job_result.JobResult:
        dto = result.to_dto()
        self._session.add(dto)
        return result

    def delete_old_entries(
        self,
        days_to_keep: value_objects.DaysToKeep,
    ) -> int:
        ts = self._ts_adapter.now().value
        cutoff: datetime.datetime = ts - datetime.timedelta(days=days_to_keep.value)
        # We need to delete batches one by one to trigger cascade deletes, a bulk update will
        # not trigger them, and we don't want to rely on specific database implementations, so
        # we cannot use ondelete='CASCADE' on the foreign key columns.
        batches: List[batch.BatchDTO] = (
            self._session.query(batch.BatchDTO).filter(batch.BatchDTO.ts < cutoff).all()
        )
        for b in batches:
            self._session.delete(b)
        return len(batches)

    def get_batch_by_id(self, batch_id: value_objects.UniqueId) -> Optional[batch.Batch]:
        result: Optional[batch.BatchDTO] = (
            self._session.query(batch.BatchDTO)
            .filter(batch.BatchDTO.id == batch_id.value)
            .first()
        )
        if result is None:
            return None
        else:
            return result.to_domain()

    def get_latest(self) -> Optional[batch.Batch]:
        result: Optional[batch.BatchDTO] = (
            self._session.query(batch.BatchDTO)
            .order_by(desc(batch.BatchDTO.ts))
            .first()
        )
        if result is None:
            return None
        else:
            return result.to_domain()

    def get_latest_result_for_job(
        self, job_name: value_objects.JobName
    ) -> Optional[job_result.JobResult]:
        dto = (
            self._session.query(job_result.JobResultDTO)
            .filter(job_result.JobResultDTO.job_name.ilike(job_name.value))  # type: ignore
            .order_by(desc(job_result.JobResultDTO.ts))
            .first()
        )
        if dto is None:
            return None
        else:
            return dto.to_domain()

    def get_last_successful_ts_for_job(
        self, job_name: value_objects.JobName
    ) -> Optional[value_objects.Timestamp]:
        jr: Optional[job_result.JobResultDTO] = (
            self._session.query(job_result.JobResultDTO)
            .filter(job_result.JobResultDTO.job_name.ilike(job_name.value))  # type: ignore
            .filter(job_result.JobResultDTO.execution_error_occurred.is_(False))  # type: ignore
            .order_by(desc(job_result.JobResultDTO.ts))
            .first()
        )
        if jr is None:
            return None
        else:
            return value_objects.Timestamp(jr.ts)

    def update(self, batch_to_update: batch.Batch) -> batch.Batch:
        dto = batch_to_update.to_dto()
        self._session.merge(dto)
        return batch_to_update



from lime_etl.adapters.timestamp_adapter import LocalTimestampAdapter, TimestampAdapter  # noqa
from lime_etl.adapters.email_adapter import DefaultEmailAdapter, EmailAdapter  # noqa
from lime_etl.domain.value_objects import *  # noqa
from lime_etl.domain.job_spec import JobSpec  # noqa
from lime_etl.domain.batch_delta import BatchDelta  # noqa
from lime_etl.domain.job_spec import AdminJobSpec, ETLJobSpec  # noqa
from lime_etl.domain.job_test_result import JobTestResult  # noqa
from lime_etl.services.unit_of_work import DefaultUnitOfWork, UnitOfWork  # noqa
from lime_etl.services.job_logging_service import DefaultJobLoggingService, JobLoggingService  # noqa
from lime_etl.services.admin.delete_old_logs import DeleteOldLogs  # noqa
from lime_etl.runner import DEFAULT_ADMIN_JOBS, run  # noqa



import typing

from lime_etl.domain import job_dependency_errors, value_objects


class LimeETLException(Exception):
    """Base class for exceptions arising from the lime-etl package"""

    def __init__(self, message: str, /):
        self.message = message
        super().__init__(message)


class DependencyErrors(LimeETLException):
    def __init__(self, dependency_errors: typing.Set[job_dependency_errors.JobDependencyErrors], /):
        self.dependency_errors = dependency_errors

        msg = "\n".join(str(e) for e in sorted(dependency_errors))
        super().__init__(msg)



import datetime
import traceback

import typing

from lime_etl.adapters import timestamp_adapter
from lime_etl.domain import job_result, job_spec, job_test_result, value_objects
from lime_etl.services import job_logging_service, unit_of_work


class JobRunner(typing.Protocol):
    def __call__(
        self,
        uow: unit_of_work.UnitOfWork,
        batch_id: value_objects.UniqueId,
        job_id: value_objects.UniqueId,
        job: job_spec.JobSpec,
        logger: job_logging_service.JobLoggingService,
        ts_adapter: timestamp_adapter.TimestampAdapter,
    ) -> job_result.JobResult:
        ...


def default_job_runner(
    uow: unit_of_work.UnitOfWork,
    batch_id: value_objects.UniqueId,
    job_id: value_objects.UniqueId,
    job: job_spec.JobSpec,
    logger: job_logging_service.JobLoggingService,
    ts_adapter: timestamp_adapter.TimestampAdapter,
) -> job_result.JobResult:
    try:
        logger.log_info(value_objects.LogMessage(f"Starting [{job.job_name.value}]..."))
        with uow:
            current_batch = uow.batches.get_batch_by_id(batch_id)

        if current_batch is None:
            raise Exception(
                "The current batch has not been saved to the database, so the results of dependent "
                "jobs cannot be checked."
            )
        else:
            dep_exceptions = {
                jr.job_name
                for jr in current_batch.job_results
                if jr.job_name in job.dependencies
                and jr.execution_success_or_failure.is_failure
            }
            dep_test_failures = {
                jr.job_name
                for jr in current_batch.job_results
                if jr.job_name in job.dependencies and jr.is_broken
            }
            if dep_exceptions and dep_test_failures:
                exceptions = ", ".join(sorted(dep_exceptions))  # type: ignore
                test_failures = ", ".join(sorted(dep_test_failures))  # type: ignore
                raise Exception(
                    f"The following dependencies failed to execute: {exceptions} "
                    f"and the following jobs had test failures: {test_failures}"
                )
            elif dep_exceptions:
                exceptions = ", ".join(sorted(dep_exceptions))  # type: ignore
                raise Exception(
                    f"The following dependencies failed to execute: {exceptions}"
                )
            else:
                result = _run_with_retry(
                    batch_id=batch_id,
                    job_id=job_id,
                    job=job,
                    logger=logger,
                    retries_so_far=0,
                    max_retries=job.max_retries.value,
                    ts_adapter=ts_adapter,
                    uow=uow,
                )
                logger.log_info(
                    value_objects.LogMessage(
                        f"Finished running [{job.job_name.value}]."
                    )
                )
                return result
    except Exception as e:
        logger.log_error(
            value_objects.LogMessage(
                (
                    f"An error occurred while running the job [{job.job_name.value}]:"
                    f"\n{traceback.format_exc(10)}"
                )
            ),
        )
        return job_result.JobResult(
            id=job_id,
            batch_id=batch_id,
            job_name=job.job_name,
            test_results=frozenset(),
            execution_success_or_failure=value_objects.Result.failure(str(e)),
            execution_millis=value_objects.ExecutionMillis(0),
            ts=ts_adapter.now(),
        )


def _run_with_retry(
    batch_id: value_objects.UniqueId,
    job_id: value_objects.UniqueId,
    job: job_spec.JobSpec,
    max_retries: int,
    retries_so_far: int,
    logger: job_logging_service.JobLoggingService,
    ts_adapter: timestamp_adapter.TimestampAdapter,
    uow: unit_of_work.UnitOfWork,
) -> job_result.JobResult:
    # noinspection PyBroadException
    try:
        return _run_job_with_tests(
            batch_id=batch_id,
            job_id=job_id,
            job=job,
            logger=logger,
            ts_adapter=ts_adapter,
            uow=uow,
        )
    except:
        if max_retries > retries_so_far:
            logger.log_info(
                value_objects.LogMessage(
                    f"Running retry {retries_so_far} or {max_retries}..."
                )
            )
            return _run_with_retry(
                batch_id=batch_id,
                job_id=job_id,
                job=job,
                max_retries=max_retries,
                retries_so_far=retries_so_far + 1,
                logger=logger,
                ts_adapter=ts_adapter,
                uow=uow,
            )
        else:
            logger.log_info(
                value_objects.LogMessage(
                    f"[{job.job_name.value}] failed after {max_retries} retries."
                )
            )
            raise


def _run_job_with_tests(
    batch_id: value_objects.UniqueId,
    job_id: value_objects.UniqueId,
    job: job_spec.JobSpec,
    logger: job_logging_service.JobLoggingService,
    ts_adapter: timestamp_adapter.TimestampAdapter,
    uow: unit_of_work.UnitOfWork,
) -> job_result.JobResult:
    start_time = datetime.datetime.now()
    if isinstance(job, job_spec.AdminJobSpec):
        result: value_objects.Result = job.run(uow=uow, logger=logger)
    elif isinstance(job, job_spec.ETLJobSpec):
        result = job.run(logger=logger)
    else:
        raise ValueError(
            f"Expected an instance of AdminJobSpec or ETLJobSpec, but got {job!r}."
        )

    if result is None:
        result = value_objects.Result.success()
    elif isinstance(result, value_objects.Result):
        logger.log_info(
            value_objects.LogMessage(f"[{job.job_name.value}] finished successfully.")
        )
    else:
        raise ValueError(
            f"A job should return an instance of either Success or Failure, but [{job.job_name.value}] "
            f"returned {result!r}."
        )

    if result.is_failure:
        logger.log_info(
            value_objects.LogMessage(
                f"[{job.job_name.value}] failed, so there is no point in running the tests.  "
                f"They have been skipped."
            )
        )
        test_results: typing.Collection[job_test_result.SimpleJobTestResult] = []
    else:
        logger.log_info(
            value_objects.LogMessage(f"Running the tests for [{job.job_name.value}]...")
        )
        if isinstance(job, job_spec.AdminJobSpec):
            test_results = job.test(logger=logger, uow=uow)
        else:
            test_results = job.test(logger=logger)  # type: ignore

    if test_results:
        full_test_results = frozenset(
            job_test_result.JobTestResult(
                id=value_objects.UniqueId.generate(),
                job_id=job_id,
                test_name=test_result.test_name,
                test_success_or_failure=test_result.test_success_or_failure,
                ts=ts_adapter.now(),
            )
            for test_result in test_results
        )
    else:
        full_test_results = frozenset()

    end_time = datetime.datetime.now()
    execution_millis = int((end_time - start_time).total_seconds() * 1000)
    ts = ts_adapter.now()
    return job_result.JobResult(
        id=job_id,
        batch_id=batch_id,
        job_name=job.job_name,
        test_results=full_test_results,
        execution_millis=value_objects.ExecutionMillis(execution_millis),
        execution_success_or_failure=value_objects.Result.success(),
        ts=ts,
    )



from __future__ import annotations

import abc

import typing

from lime_etl.domain import job_test_result, value_objects
from lime_etl.services import job_logging_service, unit_of_work


class JobSpec(abc.ABC):
    @property
    @abc.abstractmethod
    def dependencies(self) -> typing.List[value_objects.JobName]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def job_name(self) -> value_objects.JobName:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def seconds_between_refreshes(self) -> value_objects.SecondsBetweenRefreshes:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def timeout_seconds(self) -> value_objects.TimeoutSeconds:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def max_retries(self) -> value_objects.MaxRetries:
        raise NotImplementedError


class AdminJobSpec(JobSpec):
    @abc.abstractmethod
    def run(
        self,
        uow: unit_of_work.UnitOfWork,
        logger: job_logging_service.JobLoggingService,
    ) -> value_objects.Result:
        raise NotImplementedError

    @abc.abstractmethod
    def test(
        self,
        uow: unit_of_work.UnitOfWork,
        logger: job_logging_service.JobLoggingService,
    ) -> typing.Collection[job_test_result.SimpleJobTestResult]:
        raise NotImplementedError


class ETLJobSpec(JobSpec):
    @abc.abstractmethod
    def run(
        self, logger: job_logging_service.JobLoggingService
    ) -> value_objects.Result:
        raise NotImplementedError

    @abc.abstractmethod
    def test(
        self, logger: job_logging_service.JobLoggingService
    ) -> typing.Collection[job_test_result.SimpleJobTestResult]:
        raise NotImplementedError



from __future__ import annotations

import abc
import datetime

from lime_etl.domain import value_objects


class TimestampAdapter(abc.ABC):
    @abc.abstractmethod
    def now(self) -> value_objects.Timestamp:
        raise NotImplementedError


class LocalTimestampAdapter(TimestampAdapter):
    def now(self) -> value_objects.Timestamp:
        return value_objects.Timestamp(datetime.datetime.now())



import itertools
from typing import List, Protocol, runtime_checkable

import typing

from lime_etl.adapters import timestamp_adapter
from lime_etl.domain import (
    batch_delta,
    batch,
    exceptions,
    job_dependency_errors,
    job_result,
    job_spec,
    value_objects,
)
from lime_etl.services import (
    batch_logging_service,
    job_runner,
)
from lime_etl.services import job_logging_service, unit_of_work


@runtime_checkable
class BatchRunner(Protocol):
    def __call__(
        self,
        uow: unit_of_work.UnitOfWork,
        batch_id: value_objects.UniqueId,
        jobs: typing.Collection[job_spec.JobSpec],
        batch_logger: batch_logging_service.BatchLoggingService,
        ts_adapter: timestamp_adapter.TimestampAdapter,
    ) -> batch.Batch:
        ...


def run(
    uow: unit_of_work.UnitOfWork,
    jobs: typing.Collection[job_spec.JobSpec],
    ts_adapter: timestamp_adapter.TimestampAdapter,
) -> batch_delta.BatchDelta:
    batch_id = value_objects.UniqueId.generate()
    batch_logger = batch_logging_service.DefaultBatchLoggingService(
        uow=uow, batch_id=batch_id
    )
    dep_results = check_dependencies(jobs)
    if dep_results:
        e = exceptions.DependencyErrors(dep_results)

        batch_logger.log_error(
            value_objects.LogMessage("\n".join(str(e) for e in sorted(dep_results)))
        )
        raise exceptions.DependencyErrors(dep_results)

    try:
        with uow:
            previous_results = uow.batches.get_latest()
            new_batch = batch.Batch(
                id=batch_id,
                job_results=frozenset(),
                execution_millis=None,
                execution_success_or_failure=None,
                running=value_objects.Flag(True),
                ts=ts_adapter.now(),
            )
            uow.batches.add(new_batch)
            uow.commit()

        batch_logger.log_info(
            value_objects.LogMessage(f"Staring batch [{batch_id.value}]...")
        )
        result = _run_batch(
            batch_logger=batch_logger,
            uow=uow,
            batch_id=batch_id,
            jobs=jobs,
            ts_adapter=ts_adapter,
        )
        uow.batches.update(result)
        batch_logger.log_info(
            value_objects.LogMessage(f"Batch [{batch_id.value}] finished.")
        )
        return batch_delta.BatchDelta(
            current_results=result,
            previous_results=previous_results,
        )
    except Exception as e:
        batch_logger.log_error(value_objects.LogMessage(str(e)))
        raise


def check_dependencies(
    jobs: typing.Collection[job_spec.JobSpec], /
) -> typing.Set[job_dependency_errors.JobDependencyErrors]:
    job_names = {job.job_name for job in jobs}
    unresolved_dependencies_by_table = {
        job.job_name: set(dep for dep in job.dependencies if dep not in job_names)
        for job in jobs
        if any(dep not in job_names for dep in job.dependencies)
    }
    unresolved_dependencies = {
        dep for dep_grp in unresolved_dependencies_by_table.values() for dep in dep_grp
    }

    job_names_seen_so_far: typing.List[value_objects.JobName] = []
    jobs_out_of_order_by_table: typing.Dict[
        value_objects.JobName, typing.Set[value_objects.JobName]
    ] = dict()
    for job in jobs:
        job_names_seen_so_far.append(job.job_name)
        job_deps_out_of_order = []
        for dep in job.dependencies:
            if dep not in job_names_seen_so_far and dep not in unresolved_dependencies:
                job_deps_out_of_order.append(dep)
        if job_deps_out_of_order:
            jobs_out_of_order_by_table[job.job_name] = set(job_deps_out_of_order)

    return {
        job_dependency_errors.JobDependencyErrors(
            job_name=job_name,
            missing_dependencies=frozenset(unresolved_dependencies_by_table.get(job_name, set())),
            jobs_out_of_order=frozenset(jobs_out_of_order_by_table.get(job_name, set())),
        )
        for job_name in set(
            itertools.chain(
                unresolved_dependencies_by_table.keys(),
                jobs_out_of_order_by_table.keys(),
            )
        )
    }


def _run_batch(
    batch_logger: batch_logging_service.BatchLoggingService,
    uow: unit_of_work.UnitOfWork,
    batch_id: value_objects.UniqueId,
    jobs: typing.Iterable[job_spec.JobSpec],
    ts_adapter: timestamp_adapter.TimestampAdapter,
) -> batch.Batch:
    start_ts = ts_adapter.now()

    job_results: List[job_result.JobResult] = []
    for job in jobs:
        with uow:
            last_ts = uow.batches.get_last_successful_ts_for_job(job_name=job.job_name)

        if last_ts:
            seconds_since_last_refresh = (
                uow.ts_adapter.now().value - last_ts.value
            ).total_seconds()
            if seconds_since_last_refresh < job.seconds_between_refreshes.value:
                batch_logger.log_info(
                    value_objects.LogMessage(
                        f"[{job.job_name.value}] was run successfully {seconds_since_last_refresh:.0f} seconds "
                        f"ago and it is set to refresh every {job.seconds_between_refreshes.value} seconds, "
                        f"so there is no need to refresh again."
                    )
                )
                continue

        job_id = value_objects.UniqueId.generate()
        job_logger = job_logging_service.DefaultJobLoggingService(
            uow=uow, batch_id=batch_id, job_id=job_id
        )
        result = job_runner.default_job_runner(
            uow=uow,
            job=job,
            logger=job_logger,
            batch_id=batch_id,
            job_id=job_id,
            ts_adapter=ts_adapter,
        )
        job_results.append(result)
        with uow:
            uow.batches.add_job_result(result)
            uow.commit()

    end_time = ts_adapter.now().value
    execution_millis = int((end_time - start_ts.value).total_seconds() * 1000)
    return batch.Batch(
        id=batch_id,
        execution_millis=value_objects.ExecutionMillis(execution_millis),
        job_results=frozenset(job_results),
        execution_success_or_failure=value_objects.Result.success(),
        running=value_objects.Flag(False),
        ts=uow.ts_adapter.now(),
    )



import datetime
import typing
from typing import List

from lime_etl.domain import job_spec, job_test_result, value_objects
from lime_etl.services import job_logging_service, unit_of_work


class DeleteOldLogs(job_spec.AdminJobSpec):
    def __init__(
        self,
        days_to_keep: value_objects.DaysToKeep,
    ):
        self._days_to_keep = days_to_keep

    @property
    def dependencies(self) -> List[value_objects.JobName]:
        return []

    @property
    def flex_pct(self) -> value_objects.FlexPercent:
        return value_objects.FlexPercent(0)

    @property
    def max_retries(self) -> value_objects.MaxRetries:
        return value_objects.MaxRetries(1)

    @property
    def job_name(self) -> value_objects.JobName:
        return value_objects.JobName("delete_old_logs")

    @property
    def seconds_between_refreshes(self) -> value_objects.SecondsBetweenRefreshes:
        return value_objects.SecondsBetweenRefreshes(60 * 60 * 24)

    @property
    def timeout_seconds(self) -> value_objects.TimeoutSeconds:
        return value_objects.TimeoutSeconds(300)

    def run(
        self,
        uow: unit_of_work.UnitOfWork,
        logger: job_logging_service.JobLoggingService,
    ) -> value_objects.Result:
        with uow:
            uow.batch_log.delete_old_entries(days_to_keep=self._days_to_keep)
            uow.commit()

        logger.log_info(
            message=value_objects.LogMessage(
                f"Deleted batch log entries older than {self._days_to_keep.value} days old."
            )
        )

        with uow:
            uow.job_log.delete_old_entries(days_to_keep=self._days_to_keep)
            uow.commit()

        logger.log_info(
            message=value_objects.LogMessage(
                f"Deleted job log entries older than {self._days_to_keep.value} days old."
            )
        )

        with uow:
            uow.batches.delete_old_entries(days_to_keep=self._days_to_keep)
            uow.commit()

        logger.log_info(
            message=value_objects.LogMessage(
                f"Deleted batch results older than {self._days_to_keep.value} days old."
            )
        )

        return value_objects.Result.success()

    def test(
        self,
        uow: unit_of_work.UnitOfWork,
        logger: job_logging_service.JobLoggingService,
    ) -> typing.Collection[job_test_result.SimpleJobTestResult]:
        with uow:
            ts = uow.ts_adapter.now().value
            cutoff_date = datetime.datetime.combine(
                (ts - datetime.timedelta(days=self._days_to_keep.value)).date(),
                datetime.datetime.min.time(),
            )
            earliest_log_entry = uow.batch_log.get_earliest()

        if earliest_log_entry.ts.value < cutoff_date:
            return [
                job_test_result.SimpleJobTestResult(
                    test_name=value_objects.TestName(
                        "No log entries more than than 3 days old"
                    ),
                    test_success_or_failure=value_objects.Result.failure(
                        f"The earliest batch log entry is from "
                        f"{earliest_log_entry.ts.value.strftime('%Y-%m-%d %H:%M:%S')}"
                    ),
                )
            ]
        else:
            return [
                job_test_result.SimpleJobTestResult(
                    test_name=value_objects.TestName(
                        "No log entries more than than 3 days old"
                    ),
                    test_success_or_failure=value_objects.Result.success(),
                )
            ]



from __future__ import annotations

import datetime
from dataclasses import dataclass
from typing import FrozenSet, List, Optional

from lime_etl.domain import job_test_result, value_objects


@dataclass(unsafe_hash=True)
class JobResultDTO:
    id: str
    batch_id: str
    job_name: str
    test_results: List[job_test_result.JobTestResultDTO]
    execution_millis: int
    execution_error_occurred: bool
    execution_error_message: Optional[str]
    ts: datetime.datetime

    def to_domain(self) -> JobResult:
        test_results = frozenset(dto.to_domain() for dto in self.test_results)
        if self.execution_error_occurred:
            execution_success_or_failure = value_objects.Result.failure(
                self.execution_error_message or "No error message was provided."
            )
        else:
            execution_success_or_failure = value_objects.Result.success()

        return JobResult(
            id=value_objects.UniqueId(self.id),
            batch_id=value_objects.UniqueId(self.batch_id),
            job_name=value_objects.JobName(self.job_name),
            test_results=test_results,
            execution_millis=value_objects.ExecutionMillis(self.execution_millis),
            execution_success_or_failure=execution_success_or_failure,
            ts=value_objects.Timestamp(self.ts),
        )


@dataclass(frozen=True)
class JobResult:
    id: value_objects.UniqueId
    batch_id: value_objects.UniqueId
    job_name: value_objects.JobName
    test_results: FrozenSet[job_test_result.JobTestResult]
    execution_millis: value_objects.ExecutionMillis
    execution_success_or_failure: value_objects.Result
    ts: value_objects.Timestamp

    @property
    def is_broken(self) -> bool:
        return any(result.test_failed for result in self.test_results)

    def to_dto(self) -> JobResultDTO:
        test_results = [r.to_dto() for r in self.test_results]
        return JobResultDTO(
            id=self.id.value,
            batch_id=self.batch_id.value,
            job_name=self.job_name.value,
            test_results=test_results,
            execution_millis=self.execution_millis.value,
            execution_error_occurred=self.execution_success_or_failure.is_failure,
            execution_error_message=self.execution_success_or_failure.failure_message_or_none,
            ts=self.ts.value,
        )



from __future__ import annotations

import abc
from typing import Any

from sqlalchemy.orm import sessionmaker

from lime_etl.adapters import (
    batch_log_repository,
    batch_repository,
    email_adapter,
    job_log_repository,
    timestamp_adapter,
)


class UnitOfWork(abc.ABC):
    ts_adapter: timestamp_adapter.TimestampAdapter
    batches: batch_repository.BatchRepository
    batch_log: batch_log_repository.BatchLogRepository
    job_log: job_log_repository.JobLogRepository
    emailer: email_adapter.EmailAdapter

    def __enter__(self) -> UnitOfWork:
        return self

    def __exit__(self, *args: Any) -> None:
        self.rollback()

    @abc.abstractmethod
    def commit(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def rollback(self) -> None:
        raise NotImplementedError


class DefaultUnitOfWork(UnitOfWork):
    def __init__(self, session_factory: sessionmaker):
        self._session_factory = session_factory
        self.ts_adapter = timestamp_adapter.LocalTimestampAdapter()

    def __enter__(self) -> UnitOfWork:
        self._session = self._session_factory()
        self.batches = batch_repository.SqlAlchemyBatchRepository(
            session=self._session,
            ts_adapter=self.ts_adapter,
        )
        self.batch_log = batch_log_repository.SqlAlchemyBatchLogRepository(
            session=self._session,
            ts_adapter=self.ts_adapter,
        )
        self.job_log = job_log_repository.SqlAlchemyJobLogRepository(
            session=self._session,
            ts_adapter=self.ts_adapter,
        )
        return super().__enter__()

    def __exit__(self, *args: Any) -> None:
        super().__exit__(*args)
        self._session.close()

    def commit(self) -> None:
        self._session.commit()

    def rollback(self) -> None:
        self._session.rollback()



import abc
import datetime

import sqlalchemy as sa
import sqlalchemy.orm as sao

from lime_etl.adapters import timestamp_adapter
from lime_etl.domain import batch_log_entry, value_objects


class BatchLogRepository(abc.ABC):
    @abc.abstractmethod
    def add(
        self, log_entry: batch_log_entry.BatchLogEntry
    ) -> batch_log_entry.BatchLogEntry:
        raise NotImplementedError

    @abc.abstractmethod
    def delete_old_entries(self, days_to_keep: value_objects.DaysToKeep) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def get_earliest(self) -> batch_log_entry.BatchLogEntry:
        raise NotImplementedError

    @abc.abstractmethod
    def get_latest(self) -> batch_log_entry.BatchLogEntry:
        raise NotImplementedError


class SqlAlchemyBatchLogRepository(BatchLogRepository):
    def __init__(
        self,
        session: sao.Session,
        ts_adapter: timestamp_adapter.TimestampAdapter,
    ):
        self._session = session
        self._ts_adapter = ts_adapter

    def add(
        self, log_entry: batch_log_entry.BatchLogEntry
    ) -> batch_log_entry.BatchLogEntry:
        log_entry_dto = log_entry.to_dto()
        self._session.add(log_entry_dto)
        return log_entry

    def delete_old_entries(self, days_to_keep: value_objects.DaysToKeep) -> int:
        ts = self._ts_adapter.now().value
        cutoff = ts - datetime.timedelta(days=days_to_keep.value)
        return (
            self._session.query(batch_log_entry.BatchLogEntryDTO)
            .filter(batch_log_entry.BatchLogEntryDTO.ts < cutoff)
            .delete()
        )

    def get_earliest(self) -> batch_log_entry.BatchLogEntry:
        dto: batch_log_entry.BatchLogEntryDTO = (
            self._session.query(batch_log_entry.BatchLogEntryDTO)
            .order_by(batch_log_entry.BatchLogEntryDTO.ts)
            .first()
        )
        return dto.to_domain()

    def get_latest(self) -> batch_log_entry.BatchLogEntry:
        dto: batch_log_entry.BatchLogEntryDTO = (
            self._session.query(batch_log_entry.BatchLogEntryDTO)
            .order_by(sa.desc(batch_log_entry.BatchLogEntryDTO.ts))
            .first()
        )
        return dto.to_domain()



import abc
import datetime

from sqlalchemy.orm import Session

from lime_etl.adapters import timestamp_adapter
from lime_etl.domain import job_log_entry, value_objects


class JobLogRepository(abc.ABC):
    @abc.abstractmethod
    def add(
        self, log_entry: job_log_entry.JobLogEntry
    ) -> job_log_entry.JobLogEntry:
        raise NotImplementedError

    @abc.abstractmethod
    def delete_old_entries(
        self, days_to_keep: value_objects.DaysToKeep
    ) -> int:
        raise NotImplementedError


class SqlAlchemyJobLogRepository(JobLogRepository):
    def __init__(
        self, session: Session, ts_adapter: timestamp_adapter.TimestampAdapter
    ):
        self._session = session
        self._ts_adapter = ts_adapter

    def add(self, log_entry: job_log_entry.JobLogEntry) -> job_log_entry.JobLogEntry:
        log_entry_dto = log_entry.to_dto()
        self._session.add(log_entry_dto)
        return log_entry

    def delete_old_entries(
        self, days_to_keep: value_objects.DaysToKeep
    ) -> int:
        ts = self._ts_adapter.now().value
        cutoff = ts - datetime.timedelta(
            days=days_to_keep.value
        )
        return (
            self._session.query(job_log_entry.JobLogEntryDTO)
            .filter(job_log_entry.JobLogEntryDTO.ts < cutoff)
            .delete()
        )



import typing
from email.message import EmailMessage
from smtplib import SMTP

from lime_etl.domain import batch_delta, value_objects


class EmailAdapter(typing.Protocol):
    def send(self, result: batch_delta.BatchDelta) -> None:
        ...


class DefaultEmailAdapter(EmailAdapter):
    def __init__(
        self,
        smtp_server: value_objects.SMTPServer,
        smtp_port: value_objects.SMTPPort,
        username: value_objects.EmailAddress,
        password: value_objects.Password,
        recipient: value_objects.EmailAddress,
    ):
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.username = username
        self.password = password
        self.recipient = recipient

    def send(self, result: batch_delta.BatchDelta) -> None:
        if result.current_results.broken_jobs:
            if result.newly_broken_jobs:
                subject: typing.Optional[value_objects.EmailSubject] = value_objects.EmailSubject("ETL STILL BROKEN")
            else:
                subject = value_objects.EmailSubject("ETL BROKEN")
        else:
            if result.newly_fixed_jobs:
                subject = value_objects.EmailSubject("ETL FIXED")
            else:
                subject = None

        if subject:
            email_msg = value_objects.EmailMsg(
                "broken jobs: {}".format(
                    ", ".join(jn.value for jn in result.current_results.broken_jobs)
                )
                + "newly broken jobs: {}".format(
                    ", ".join(jn.value for jn in result.newly_broken_jobs)
                )
                + "newly fixed jobs: {}".format(
                    ", ".join(jn.value for jn in result.newly_fixed_jobs)
                )
            )
            with SMTP(host=self.smtp_server.value, port=self.smtp_port.value) as s:
                s.login(user=self.username.value, password=self.password.value)
                msg = EmailMessage()
                msg.set_content(email_msg.value)
                msg["Subject"] = subject.value
                msg["From"] = self.username.value
                msg["To"] = self.recipient.value
                s.send_message(msg)
                s.quit()
                return None
        else:
            return None
